# -*- coding: utf-8 -*-
"""
Created on Wed Dec  9 09:39:53 2020

@author: Lucia
"""

import pandas as pd
import numpy as np

from sklearn.cluster import OPTICS

from sklearn.metrics import pairwise_distances

from sklearn.metrics import silhouette_score
from sklearn.metrics import calinski_harabasz_score
from sklearn.metrics import davies_bouldin_score

import matplotlib.pyplot as plt


#------------------------------------------------------------------------------
def do_clustering(ms, dist_matrix):

    # Generamos el modelo a partir de los parámetros
    labels = OPTICS(min_samples = ms, 
                    metric = "precomputed"
                    ).fit_predict(dist_matrix)

    # Todos los elementos son asignados como "-1" ó como "0"
    if len(np.unique(labels)) < 2:
        n_clusters = 0 if -1 in labels else 1
        n_noise = len(labels) if -1 in labels else 0
        silh_coef = -1
        cal_hara_coef = 1
        db_coef = 0

    else:
        n_clusters = len(set(labels)) - (1 if -1 in labels else 0)
        n_noise = list(labels).count(-1)
        silh_coef = silhouette_score(dist_matrix,
                                      labels,
                                      metric = 'precomputed')
        cal_hara_coef = calinski_harabasz_score(dist_matrix,
                                                labels)
        db_coef = davies_bouldin_score(dist_matrix,
                                        labels)
                
    return(n_clusters, n_noise, silh_coef, cal_hara_coef, db_coef)
#------------------------------------------------------------------------------





#------------------------------------------------------------------------------
def do_clustering_best_model(dataset, vector, ind, feature):
    
    if vector == 'bool':
        v = 'binary'
     
    else:
         v = 'numeric'
    
    dist_matrix = np.load('data/distances/' + dataset + '/' + vector + '/' + ind + '_' + feature + '.npy')
    feat_dis = pd.read_csv('data/disease_feature/' + dataset + '/' + vector + '/disease_' + feature + '.csv')

    
    diseases = feat_dis[['disease_id']].drop_duplicates()
    
    # Generamos el modelo a partir de los parámetros
    # clusterer = DBSCAN(eps = 0.4, min_samples = 2, metric = 'precomputed')
    clusterer = OPTICS(min_samples = 2,
                       metric = 'precomputed')
    clusterer.fit(dist_matrix)
    
    diseases[['cluster']] = clusterer.labels_
    diseases.reset_index(drop=True, inplace=True)
    
    clusters = pd.DataFrame()
    clusters[['cluster']] = diseases[['cluster']].drop_duplicates()
    clusters = clusters.sort_values(by = ['cluster'])
    clusters.drop(clusters[clusters['cluster'] == -1].index, inplace=True)
    
    numbers = list()   
    for cluster in clusters.values.tolist():
        numbers.append(len(diseases[diseases['cluster'] == cluster[0]]))
        
    clusters[['numbers']] = numbers
    clusters.reset_index(drop=True, inplace=True)
    


    plt.figure(figsize=(20, 15))
    clusters.hist(column = 'numbers', 
                       bins = 30, 
                        grid = False,
                       # range = (0,35),
                       align = 'left',
                       rwidth=0.8,
                       color='#607c8e')
    plt.title("Clustering model with OPTICS", fontweight='bold')
    plt.xlabel("Number of diseases inside the cluster")
    plt.ylabel("Number of clusters")
    props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)
    plt.text(25,
             100, 
             'Dataset: '+ dataset +'\nVectors: ' + v + '\nDistance: '+ind+'\nFeature: '+feature+'s',
             bbox=props)
    plt.gca().spines['right'].set_color('none')
    plt.gca().spines['top'].set_color('none')
    plt.show()
    
    return(diseases, clusters)
#------------------------------------------------------------------------------





#------------------------------------------------------------------------------
def do_optimization(dist_matrix, ind, ms_values):
    
    # Listas que almacenan la combinación de parámetros y los resultados
    indss = list() # Índices de similitud
    mss = list() # Valores de MinPts
    clusters = list() # Resultados del número de clusters
    noise = list() # Resultados del número de outliers
    silhouettes = list() # Resultados del coeficiente de silhouette
    calinskis = list() # Resultados del score de calinski-harabasz
    dbs = list() # Resultados del score de davies-bouldin
    
    # Recorremos los 4 posibles valores de MinPts
    for ms in ms_values:    
           
            n_clusters, n_noise, silh_coef, cal_hara_coef, db_coef = do_clustering(ms, dist_matrix)
            
            # Guardamos los parámetros y los resultados
            indss.append(ind)
            mss.append(ms)
            clusters.append(n_clusters)
            noise.append(n_noise)
            silhouettes.append(silh_coef)
            calinskis.append(cal_hara_coef)
            dbs.append(db_coef)
    
    results = pd.DataFrame(list(zip(indss,
                                    mss,
                                    clusters,
                                    noise,
                                    silhouettes,
                                    calinskis,
                                    dbs)),
                            columns = ['Similarity',
                                       'MinPts',
                                       'Cluster',
                                       'Noise',
                                       'Silhouette',
                                       'Calinski-Harabasz',
                                       'Davies-Bouldin'])
    
    return(results)
#------------------------------------------------------------------------------





#------------------------------------------------------------------------------
def simluations(n_simulations, original, index, ms, feature, vector, dataset):
    
    original_np = original.to_numpy()
    
    clusters = list()
    noise = list()
    silhouettes = list()
    calinskis = list() 
    dbs = list() 
    
    for n in range(n_simulations):
        
        shuffled = np.copy(original_np)
    
        # Shuffling elements in each column of the matrix
        for i in range(shuffled.shape[1]):
            np.random.shuffle(shuffled[:,i])
    
        distances = pairwise_distances(shuffled,
                                       metric = index)
    
        n_clusters, n_noise, silh_coef, cal_hara_coef, db_coef = do_clustering(ms, distances)
    
        clusters.append(n_clusters)
        noise.append(n_noise)
        silhouettes.append(silh_coef)
        calinskis.append(cal_hara_coef)
        dbs.append(db_coef)
    
    
    results = pd.DataFrame(list(zip(clusters, noise, silhouettes, calinskis, dbs)),
                          columns = ['Clusters', 'Noise', 'Silhouette', 'Calinski-Harabasz', 'Davies-Bouldin'])
    
    
    results.to_csv("results/simulations/optics/" + dataset + '/' + feature + "s_" + vector + '_' + index + "_ms" + str(ms) + ".csv", index=False)
#------------------------------------------------------------------------------